import os
import unittest
import atooms.trajectory
import berni
try:
    import rumd
    import berni.rumd
    from atooms.backends.rumd import RUMD
    SKIP = False
except ImportError:
    SKIP = True

class Test(unittest.TestCase):

    def setUp(self):
        if SKIP:
            self.skipTest('missing RUMD')
        self.root = os.path.dirname(__file__)

    def test_(self):
        for model in ['kob_andersen', 'coslovich_pastore']:
            pot = berni.rumd.potential(model)

    def test_cp(self):
        import berni
        inp = berni.samples.get('coslovich_pastore-488db481cdac35e599922a26129c3e35.xyz')
        atooms.trajectory.TrajectoryXYZ(inp).copy(cls=atooms.trajectory.TrajectoryRUMD, fout='/tmp/1.xyz')
        model = 'coslovich_pastore'
        potentials = berni.rumd.potential(model)
        backend = RUMD('/tmp/1.xyz', potentials=potentials)
        epot_0 = backend.system.potential_energy(per_particle=True)

        model = 'coslovich_pastore'
        model = berni.get(model)
        model['cutoff'][0]['type'] = 'linear_cut_shift'
        model['cutoff'][0]['parameters']['rcut'] = model['cutoff'][0]['parameters']['rspl']
        model['cutoff'][0]['parameters'].pop('rspl')
        with atooms.trajectory.Trajectory(inp) as th:
            s = th[0]
            s.species_layout = 'F'
            s.interaction = berni.f90.Interaction(model)
            epot_1 = s.potential_energy(per_particle=True)
        self.assertLess(abs(epot_0 - epot_1), 1e-6)
        import os
        os.system(f'rm -f {inp}')

    def test_ka(self):
        inp = berni.samples.get('kob_andersen-8f4a9fe755e5c1966c10b50c9a53e6bf.xyz')
        atooms.trajectory.TrajectoryXYZ(inp).copy(cls=atooms.trajectory.TrajectoryRUMD, fout='/tmp/1.xyz')
        model = 'kob_andersen'
        potentials = berni.rumd.potential(model)
        backend = RUMD('/tmp/1.xyz', potentials=potentials)
        epot_0 = backend.system.potential_energy(per_particle=True)
        model = 'kob_andersen'
        with atooms.trajectory.Trajectory(inp) as th:
            s = th[0]
            s.species_layout = 'F'
            s.interaction = berni.f90.Interaction(model)
            epot_1 = s.potential_energy(per_particle=True)
        self.assertLess(abs(epot_0 - epot_1), 1e-6)
        import os
        os.system(f'rm -f {inp}')

    def tearDown(self):
        import os
        os.system('rm -f /tmp/1.xyz')


if __name__ == '__main__':
    unittest.main()



import os
import json
from . import models


def export(model, *args, **kwargs):
    """
    Export `model` as an atooms.f90.Interaction object, passing the
    remaining arguments to the Interaction constructor
    """
    try:
        from atooms.backends.f90 import Interaction
    except ImportError:
        print('atooms is not installed')
        raise

    if not hasattr(model, 'get'):
        if os.path.isfile(model) and model.endswith('json'):
            # This is a json file, we read it
            with open(model) as fh:
                model = json.load(fh)
        else:
            # This may be a string, so we look for the model in the
            # database and replace the string with the dictionary
            model = models.get(model)

    return Interaction(model, *args, **kwargs)


Interaction = export



"""
Database of interaction models for classical molecular dynamics
and Monte Carlo simulations
"""

import os
import glob
import json
import tempfile
import shutil
from copy import deepcopy
from tinydb import Query
from tinydb import TinyDB as _TinyDB
from tinydb.storages import MemoryStorage
from .helpers import _wget, pprint
from .schema import schema_version
from . import _schemas


class TinyDB(_TinyDB):

    default_storage_class = MemoryStorage
    default_table_name = 1
    _get_payload = None
    _str_kwargs = {'include': ['name']}

    def get(self, name, **kwargs):
        """Get the payload associated to the qualified name"""
        return self._get_payload(self, name, **kwargs)

    def fields(self, merge=set.union):
        """Return fields of database"""
        cols = None
        for entry in self:
            if cols is None:
                cols = set(entry.keys())
            else:
                cols = merge(cols, set(entry.keys()))
        return sorted(list(cols))

    def pprint(self, include=None, cond=None, sort_by=None,
               max_rows=20, max_len=140, file=None):
        kwargs = locals()
        kwargs.pop('cond')
        kwargs.pop('self')
        if cond is None:
            pprint(self.all(), **kwargs)
        else:
            pprint(self.search(cond), **kwargs)

    def __str__(self):
        from .helpers import pprint
        from io import StringIO
        with StringIO() as io:
            pprint(self.all(), **self._str_kwargs, file=io)
            txt = io.getvalue()
        return txt

def _get_model(db, name, schema_version=1):
    """
    Return the model matching the qualified name  `name`

    The qualified name has the form <model_name>[-<variant>]
    """
    # Look for a single match
    query = Query()
    db.default_table_name = schema_version
    entry = db.search(query.name == name)
    if len(entry) == 1:
        # Get the actual payload
        return deepcopy(entry[0]['_model'])
    if len(entry) == 0:
        raise KeyError(f'Model {name} not found with schema {schema_version}')
    if len(entry) > 1:
        raise KeyError(f'Multiple models {name} found')

def _get_sample(db, name, output_path=None):
    """
    Get a copy of sample `name` and return the actual path to it
    """
    query = Query()
    entry = db.search((query.name == name))[0]

    if output_path is None:
        tmpdir = tempfile.mkdtemp()
        output_path = os.path.join(tmpdir, name)

    if 'url' in entry:
        _wget(entry['url'], tmpdir)
    else:
        url = os.path.join(_samples, entry['name'])
        shutil.copy(url, output_path)
    return output_path


_root = os.path.dirname(__file__)
_samples = os.path.join(_root, 'samples')
schemas = {1: _schemas.m1, 2: _schemas.m2}
models = TinyDB()
samples = TinyDB()
models._str_kwargs = {'include': ['name', 'reference'], 'sort_by': 'name'}
samples._str_kwargs = {'include': ['name', 'model'], 'sort_by': 'name'}
# Monkey patch TinyDBs with custom getters
models._get_payload = _get_model
samples._get_payload = _get_sample

# Aliases (get and model) for backward compatibility
def model(model, schema_version=1):
    return models.get(model, schema_version=schema_version)


get = model

# Store models
def _store_models(models):
    """Store models in the global database"""
    for f in glob.glob(f'{_root}/models/*.json'):
        with open(f) as fh:
            model = json.load(fh)
        sv = schema_version(model)
        # if 'variant' not in model['metadata']:
        #     model['metadata']['variant'] = 0
        # Store model in database table of this schema version
        table = models.table(sv)
        metadata = model.pop('metadata')
        table.insert({**metadata, '_model': model})

# Store samples
def _store_samples(samples):
    """Store samples in the global database"""
    for path in glob.glob(f'{_samples}/*.json'):
        with open(path) as fh:
            data = json.load(fh)
            samples.insert(data)


_store_models(models)
_store_samples(samples)


# Convenience export function for backends
def export(model, backend, *args, **kwargs):
    """
    Export `model` to suitable object for `backend`, passing
    additional optional arguments to the backend
    """
    from . import f90, rumd
    _map = {
        'f90': f90,
        'rumd': rumd
    }
    return _map[backend].export(model, *args, **kwargs)


# Potentials and cutoffs
from inspect import getmembers, isfunction, isclass
from .helpers import _objdict
from . import _potentials
from . import _cutoffs

potentials = _objdict()
for name, func in getmembers(_potentials, isfunction):
    potentials[name] = func

cutoffs = _objdict()
for name, cls in getmembers(_cutoffs, isclass):
    if name != '_objdict':
        cutoffs[name] = cls

def potential(name):
    return potentials[name]

def cutoff(name):
    return cutoffs[name]



import os
import shutil
import hashlib
from .schema import schema_version


def pprint(rows, include=None, sort_by=None, max_rows=20, max_len=140, file=None):
    """Pretty print `rows` (a list of dicts)"""

    def _tabular(data, max_len=max_len):
        """General function to format `data` list in tabular table"""
        # Predict formatting
        lens = [0 for _ in range(len(data[0]))]
        for entry in data:
            for i, value in enumerate(entry):
                lens[i] = max(lens[i], len(str(value)))
        fmts = [f'{{:{lens[i]}s}}' for i in range(len(lens))]
        fmt = ' '.join(fmts)

        # Store list of lines
        lines = []
        lines.append(fmt.format(*data[0]))
        lines.append('-'*(sum(lens) + len(lens) - 1))
        for entry in data[1:]:
            entry = [str(_) for _ in entry]
            lines.append(fmt.format(*entry))
            if len(lines) > max_rows and max_rows > 0:
                lines.append(f'... {len(data) - max_rows} entries not shown')
                break

        # Limit columns
        if sum(lens) > max_len:
            for i, line in enumerate(lines):
                if i < 2:
                    fill = '     '
                else:
                    fill = ' ... '
                lines[i] = line[:max_len//2] + fill + line[sum(lens) - max_len//2:]
        return lines

    # Format and sort the data
    if include is None:
        columns = set([e for e in rows[0] if not e.startswith('_')])
        for entry in rows:
            new_columns = set([e for e in entry if not e.startswith('_')])
            columns = set.union(columns, new_columns)
        columns = sorted(columns)
    else:
        columns = include

    if sort_by is not None:
        if not (isinstance(sort_by, list) or isinstance(sort_by, tuple)):
            sort_by = [sort_by]
        rows = sorted(rows[1:], key=lambda x: [x[_] for _ in sort_by])

    # Tabularize lines and join them
    rows = [columns] + [[str(entry.get(key)) for key in columns] for entry in rows]
    lines = _tabular(rows)
    print('\n'.join(lines), file=file)

def _wget(url, output_dir):
    """Like wget on the command line"""
    try:
        from urllib.request import urlopen  # Python 3
    except ImportError:
        from urllib2 import urlopen  # Python 2

    basename = os.path.basename(url)
    output_file = os.path.join(output_dir, basename)
    response = urlopen(url)
    length = 16*1024
    with open(output_file, 'wb') as fh:
        shutil.copyfileobj(response, fh, length)

class _objdict(dict):

    """Boots a dict with object-like attribute accessor"""

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

# TODO: clean up and restore the functions below, for internal use

# To add a model
# def add_model_json(path):
#     """
#     If `path` is a directory, add all json files in there to the
#     global `database`. If `path` ends with `json`, it will be assumed
#     to be match one or multiple json files (ex. `*.json`).
#     """
#     import json
#     import glob
#     if path.endswith('json'):
#         search_path = glob.glob(path)
#     else:
#         search_path = glob.glob('{}/*.json'.format(path))

#     for _path in search_path:
#         # Read json file
#         with open(_path) as fh:
#             try:
#                 model = json.load(fh)
#             except (ValueError, json.decoder.JSONDecodeError):
#                 print('Error reading file {}'.format(_path))
#                 raise

#         # By default, the model name is the file basename (stripped of .json)
#         if 'name' not in model:
#             name = os.path.basename(_path)[:-5]
#             model['name'] = '-'.join([entry.capitalize() for entry in name.split('_')])

#         yield model

# To add a sample
# def update(pretend=False):
#     for sample in samples():
#         model, path = sample["model"], sample["path"]
#         sample = {k: sample[k] for k in sample if k not in ["model", "path"]}
#         if path.startswith('http'):
#             store(path, model, pretend=pretend, **sample)
#         else:
#             store(os.path.join(_storage, path), model, pretend=pretend, **sample)


# def _store(path, model, version=0, format=None, notes="", state="", pretend=False, **kwargs):
#     """
#     Keyword arguments can be added to describe the thermodynamic state
#     of the system.
#     """
#     import tempfile
#     _storage_path = '{model}-{variant}-{md5_hash}'

#     # Set paths: storage_path is what goes in the db
#     if path.startswith('http'):
#         tmpdir = tempfile.mkdtemp()
#         basename = os.path.basename(path)
#         _wget(path, tmpdir)
#         local_path = os.path.join(tmpdir, basename)
#     else:
#         local_path = path

#     # We now have a local copy in local_path, get the md5 hash
#     with open(local_path, "rb") as fh:
#         data = fh.read()
#         md5_hash = hashlib.md5(data).hexdigest()
#         extension = os.path.splitext(local_path)[-1]

#     _locals = {}
#     _locals.update(kwargs)
#     _locals['md5_hash'] = md5_hash
#     _locals['model'] = model
#     _locals['version'] = version
#     storage_path = _storage_path.format(**_locals)

#     # Add the entry and its metadata
#     entry = {}
#     entry.update(**kwargs)
#     entry.update({
# 	"model": model,
#         "version": version,
#         "md5_hash": md5_hash,
#         "path": path,
# 	"format": format,
#     })
#     if not path.startswith('http'):
#         entry["path"] = storage_path + extension

#     if pretend:
#         print(os.path.join(_storage, storage_path))
#         return

#     from atooms.core.utils import mkdir
#     mkdir(os.path.dirname(os.path.join(_storage, storage_path)))

#     # Store json file
#     with open(os.path.join(_storage, storage_path + ".json"), 'w') as f:
#         json.dump(entry, f, indent=4)

#     # Store the json file in the storage/ (or samples/).
#     # I would like to use a descriptive path name, but it could be
#     # that the kwargs are not enough to distinguish them.
#     # So I use the md5 hash itself, but prefixed with model name and
#     # version, to have ar least a feeling of what is it.

#     # TODO: issue when having two entries that point to the same file,
#     # but one is a local file reference (within the repo) and the
#     # other one is via the url (on framagit).
#     # But in principle, one could have multiple ways to get to a sample (for redundancy)
#     # If they are all mapped to the same .json file, it does not work.
#     # 1) add a warning that we are overwriting
#     # 2) perhaps add some reference in the path to the storage (file or http)
#     # Actually, there could be multiple instances over the network,
#     # one would need to hash the url!
#     if not path.startswith('http'):
#         with open(os.path.join(_storage, entry["path"]), "wb") as fh:
#             fh.write(data)



from math import exp, log

__all__ = ['inverse_power', 'sum_inverse_power', 'yukawa', 'lennard_jones', 'fene', 'gaussian', 'harmonic']

def inverse_power(r, epsilon=1.0, sigma=1.0, exponent=12):
    return \
        epsilon*(sigma/r)**exponent, \
        epsilon*exponent*(sigma/r)**exponent/r**2, \
        (-epsilon*exponent**2*(sigma/r)**exponent/r**2 - epsilon*exponent*(sigma/r)**exponent/r**2)/r

def sum_inverse_power(r, epsilon=(1.0, ), sigma=(1.0, ), exponent=(12, )):
    return 0.0, 0.0, 0.0

def yukawa(r, epsilon=1.0, kappa=1.0, sigma=1.0):
    return \
        epsilon*exp(kappa*(-r + sigma))/(kappa*r), \
        epsilon*(kappa*r + 1)*exp(-kappa*(r - sigma))/(kappa*r**3), \
        epsilon*(-kappa**2*r**2 - 2*kappa*r - 2)*exp(-kappa*(r - sigma))/(kappa*r**4)

def lennard_jones(r, epsilon=1.0, sigma=1.0):
    return \
        4*epsilon*(-sigma**6/r**6 + sigma**12/r**12), \
        -4*epsilon*(6*sigma**6/r**7 - 12*sigma**12/r**13)/r, \
        -4*epsilon*(-42*sigma**6/r**8 + 156*sigma**12/r**14)/r

def fene(r, epsilon=1.0, kappa=1.0, R=1.0, sigma=1.0):
    u = -R**2*kappa*log(-r**2/sigma**2 + 1)
    w = 2*R**2*kappa/(r**2 - sigma**2)
    h = 2*R**2*kappa*(-r**2 - sigma**2)/(r*(r**2 - sigma**2)**2)
    if (r**2 < 2**(1./6)):
        u += 4*epsilon*(-sigma**6/r**6 + sigma**12/r**12) + epsilon
        w += 24*epsilon*sigma**6*(-r**6 + 2*sigma**6)/r**14
        h += 24*epsilon*sigma**6*(7*r**6 - 26*sigma**6)/r**15
    return u, w, h

def gaussian(r, epsilon=1.0, sigma=1.0):
    return \
        epsilon*exp(-r**2/sigma**2), \
        2*epsilon*exp(-r**2/sigma**2)/sigma**2, \
        2*epsilon*(-2*r**2 + sigma**2)*exp(-r**2/sigma**2)/(r*sigma**4)
    return 0.0

def harmonic(r, epsilon=1.0, sigma=1.0):
    return \
        0.5*epsilon*(-r/sigma + 1)**2, \
        1.0*epsilon*(-r + sigma)/(r*sigma**2), \
        -1.0*epsilon/(r*sigma**2)



import numpy as np
from .helpers import _objdict

__all__ = ['shift', 'cut', 'linear', 'quadratic',
           'quadratic_cut_shift', 'linear_cut_shift', 'cut_shift']

class shift:

    def __init__(self, func, params, cutoff_params):
        nsp = cutoff_params['rcut'].shape[0]
        cutoff_params['shift'] = np.ndarray((nsp, nsp))
        for isp in range(nsp):
            for jsp in range(nsp):
                cutoff_params['shift'][isp, jsp] = func(cutoff_params['rcut'][isp, jsp], isp, jsp, **params)
        # Somehow jax only accepts a reference to a dict
        self._params = cutoff_params

    def __call__(self, func):
        def wrapper(r, isp, jsp, *args, **kwargs):
            return func(r, isp, jsp, *args, **kwargs) - self._params['shift'][isp, jsp]
        return wrapper

class cut:
    pass

class linear:
    pass

class quadratic:
    pass

class cubic_spline:
    pass


cut_shift = shift
linear_cut_shift = linear
quadratic_cut_shift = quadratic



import os
import unittest
from berni import models, samples
from atooms.trajectory import Trajectory
from atooms.system import System, Particle, Cell
from berni.f90 import Interaction


class Test(unittest.TestCase):

    def setUp(self):
        self.fileinp = os.path.join(os.path.dirname(__file__), '../berni/samples/lennard_jones-13ce47602b259f7802e89e23ffd57f19.xyz')
        self.trajectory = Trajectory(self.fileinp)

    def test_collinear(self):
        model = models.get("lennard_jones")
        particles = [Particle(position=[0.0, 0.0, 0.0], species=1),
                     Particle(position=[1.0, 0.0, 0.0], species=1),
                     Particle(position=[2.0, 0.0, 0.0], species=1)]
        cell = Cell([10., 10., 10.])
        system = System(particles, cell)
        system.interaction = Interaction(model)
        self.assertAlmostEqual(system.potential_energy(), -0.01257276409199999)

    def test_ntw(self):
        trajectory = Trajectory(os.path.join(os.path.dirname(__file__), '../berni/samples/coslovich_pastore-488db481cdac35e599922a26129c3e35.xyz'))
        system = trajectory[0]
        system.species_layout = 'F'
        system.interaction = Interaction('coslovich_pastore')
        self.assertAlmostEqual(system.potential_energy(per_particle=True), -6.0295, places=4)

    def tearDown(self):
        self.trajectory.close()


if __name__ == '__main__':
    unittest.main()
